from __future__ import annotations

import asyncio
import hashlib
import pathlib
from typing import TYPE_CHECKING

import aiopath
from watchfiles import Change, awatch

from pylav.logging import getLogger
from pylav.nodes.api.responses.rest_api import SearchResult, TrackLoaded
from pylav.nodes.api.responses.track import Track
from pylav.players.query.local_files import ALL_EXTENSIONS
from pylav.players.query.obj import Query

if TYPE_CHECKING:
    from pylav.core.client import Client

LOGGER = getLogger("PyLav.LocalTrackCache")


class LocalTrackCache:
    __slots__ = (
        "__pylav",
        "__ready",
        "__monitor",
        "__shutdown",
        "__query_cache",
        "__track_cache",
        "__root_folder",
        "__query_lock",
        "__track_lock",
        "__path_to_query_cache",
    )

    def __init__(self, client: Client, root: str | pathlib.Path | aiopath.Path) -> None:
        self.__shutdown = False
        self.__pylav = client
        self.__query_cache: dict[str, Query] = {}
        self.__track_cache: dict[str, Track] = {}
        self.__path_to_query_cache: dict[str, Query] = {}
        self.__query_lock = asyncio.Lock()
        self.__track_lock = asyncio.Lock()
        self.__root_folder = pathlib.Path(root)
        self.__ready = asyncio.Event()
        self.__monitor = asyncio.create_task(self.file_watcher())

    def __bool__(self) -> bool:
        return not self.__shutdown

    @property
    def hexdigest_to_query(self) -> dict[str, Query]:
        return self.__query_cache

    @property
    def path_to_query(self) -> dict[str, Query]:
        return self.__path_to_query_cache

    @property
    def path_to_track(self) -> dict[str, Track]:
        return self.__track_cache

    @property
    def root_folder(self) -> pathlib.Path:
        return self.__root_folder

    @property
    def is_ready(self) -> bool:
        return self.__ready.is_set()

    async def initialize(self):
        await self.__pylav.wait_until_ready()
        await self.update()
        self.__ready.set()

    async def shutdown(self):
        self.__shutdown = True
        self.__monitor.cancel()
        self.__ready.clear()
        await self.wipe_cache()

    async def _add_to_query_cache(self, query: Query, path: str) -> None:
        if self.__shutdown:
            return
        async with self.__query_lock:
            self.__query_cache[hashlib.md5(f"{query._query}".encode()).hexdigest()] = query
            self.__path_to_query_cache[path] = query

    async def _remove_from_query_cache(self, query: Query, path: str) -> None:
        if self.__shutdown:
            return
        async with self.__query_lock:
            self.__query_cache.pop(hashlib.md5(f"{query._query}".encode()).hexdigest(), None)
            self.__path_to_query_cache.pop(path, None)

    async def _add_to_track_cache(self, track: Track, path: pathlib.Path) -> None:
        if self.__shutdown:
            return
        async with self.__track_lock:
            self.__track_cache[f"{path}"] = track

    async def _remove_from_track_cache(self, path: pathlib.Path) -> None:
        if self.__shutdown:
            return
        async with self.__track_lock:
            self.__track_cache.pop(f"{path}", None)

    async def wipe_cache(self) -> None:
        await self.__track_lock.acquire()
        await self.__query_lock.acquire()
        self.__track_cache.clear()
        self.__query_cache.clear()
        self.__path_to_query_cache.clear()
        self.__track_lock.release()
        self.__query_lock.release()

    async def file_watcher(self):
        await self.__ready.wait()
        async for changes in awatch(self.root_folder, recursive=True):
            if self.__shutdown:
                return
            await self._process_changes(changes)

    async def _process_changes(self, changes: set[tuple[Change, str]]) -> None:
        for change, path in changes:
            path_obj = pathlib.Path(path)
            if (not path_obj.is_dir()) and path_obj.suffix.lower() not in ALL_EXTENSIONS:
                continue
            if change == Change.added:
                await self._process_added(path, path_obj)
                LOGGER.verbose(f"Added {path}")
            elif change == Change.modified:
                await self._process_modified(path, path_obj, modified=True)
                LOGGER.verbose(f"Modified {path}")
            elif change == Change.deleted:
                await self._process_deleted(path, path_obj)
                LOGGER.verbose(f"Deleted {path}")

    async def _process_added(self, path: str, path_obj: pathlib.Path, modified: bool = False) -> None:
        if self.__shutdown:
            return
        query = await Query.from_string(path_obj)
        if path_obj.is_dir():
            await self._add_to_query_cache(query, path)
            return
        await asyncio.sleep(0.1)
        track = await self.__pylav.get_tracks(query, bypass_cache=modified)
        if isinstance(track, (TrackLoaded, SearchResult)):
            await self._add_to_query_cache(query, path)
            await self._add_to_track_cache(track.tracks[0], path)

    async def _process_modified(self, path: str, path_obj: pathlib.Path, modified: bool = True) -> None:
        if self.__shutdown:
            return
        query = await Query.from_string(path_obj)
        await self._remove_from_query_cache(query, path)
        await self._remove_from_track_cache(path)
        if path_obj.is_dir():
            await self._add_to_query_cache(query, path)
            return
        await asyncio.sleep(0.1)
        track = await self.__pylav.search_query(query, bypass_cache=modified)
        if isinstance(track, (TrackLoaded, SearchResult)):
            await self._add_to_query_cache(query, path)
            await self._add_to_track_cache(track.tracks[0], path)

    async def _process_deleted(self, path: str, path_obj: pathlib.Path) -> None:
        if self.__shutdown:
            return
        query = await Query.from_string(path_obj)
        await self._remove_from_query_cache(query, path)
        await self._remove_from_track_cache(path)

    async def update(self) -> None:
        if self.__shutdown:
            return
        await self.__pylav.wait_until_ready()
        LOGGER.debug("Updating cache")
        for entry in self.__root_folder.rglob("*"):
            if (not entry.is_dir()) and entry.suffix.lower() not in ALL_EXTENSIONS:
                continue
            await self._process_added(f"{entry}", entry)
        LOGGER.debug("Finished updating cache")



from __future__ import annotations

import dataclasses
from typing import Literal


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class LoadException:
    severity: Literal["COMMON", "SUSPICIOUS", "FAULT"]
    message: str | None = None


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class LavalinkException(LoadException):
    cause: str | None = None  # This is only optional so that inheritance in python works



from __future__ import annotations

import asyncio
from pathlib import Path

from redbot.core.i18n import Translator

from pylav.core.context import PyLavContext
from pylav.extension.red.ui.menus.generic import EntryPickerMenu
from pylav.extension.red.ui.selectors.generic import EntrySelectSelector
from pylav.extension.red.ui.sources.generic import EntryPickerSource
from pylav.helpers.format.strings import shorten_string
from pylav.type_hints.bot import DISCORD_COG_TYPE
from pylav.type_hints.generics import ANY_GENERIC_TYPE

_ = Translator("PyLav", Path(__file__))


async def maybe_prompt_for_entry(
    cog: DISCORD_COG_TYPE, context: PyLavContext, entries: list[ANY_GENERIC_TYPE], message_str: str, selector_text: str
) -> ANY_GENERIC_TYPE | None:
    """Prompt the user to pick an item from a list of items.

    Ideally the item objects should have a `.name` attribute and a `.id` attribute.

    If a name is not available, this uses the class name which may result is duplicate entries.
    if no id is available we use the name hash.

    if no entry is selected or user closes the menu, None is returned.

    """
    if len(entries) > 1:
        entry_picker = EntryPickerMenu(
            cog=cog,
            bot=cog.bot,
            source=EntryPickerSource(
                guild_id=context.guild.id,
                cog=cog,
                pages=entries,
                message_str=message_str,
            ),
            selector_cls=EntrySelectSelector,
            delete_after_timeout=True,
            clear_buttons_after=True,
            starting_page=0,
            selector_text=shorten_string(max_length=100, string=selector_text),
            original_author=context.interaction.user if context.interaction else context.author,
        )

        await entry_picker.start(context)
        try:
            await entry_picker.wait_for_response()
            entry = entry_picker.result
        except asyncio.TimeoutError:
            entry = None
    else:
        entry = entries[0]
    return entry



from __future__ import annotations

import dataclasses
from typing import Annotated

from pylav.nodes.api.responses.filters.misc import ValueRange


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class ChannelMix:
    leftToLeft: Annotated[float | None, ValueRange(min=0.0, max=1.0)] = None
    leftToRight: Annotated[float | None, ValueRange(min=0.0, max=1.0)] = None
    rightToLeft: Annotated[float | None, ValueRange(min=0.0, max=1.0)] = None
    rightToRight: Annotated[float | None, ValueRange(min=0.0, max=1.0)] = None

    def to_dict(self) -> dict[str, float]:
        response = {}
        if self.leftToLeft is not None:
            response["leftToLeft"] = self.leftToLeft
        if self.leftToRight is not None:
            response["leftToRight"] = self.leftToRight
        if self.rightToLeft is not None:
            response["rightToLeft"] = self.rightToLeft
        if self.rightToRight is not None:
            response["rightToRight"] = self.rightToRight
        return response



from __future__ import annotations

from pylav.players.filters.misc import FilterMixin


class Karaoke(FilterMixin):
    __slots__ = ("_level", "_mono_level", "_filter_band", "_filter_width", "_default")

    def __init__(
        self,
        level: float | None = None,
        mono_level: float | None = None,
        filter_band: float | None = None,
        filter_width: float | None = None,
    ) -> None:
        super().__init__()
        self.level = level
        self.mono_level = mono_level
        self.filter_band = filter_band
        self.filter_width = filter_width

    def to_dict(self) -> dict[str, float | bool | None]:
        return {
            "level": self.level,
            "mono_level": self.mono_level,
            "filter_band": self.filter_band,
            "filter_width": self.filter_width,
        }

    @classmethod
    def from_dict(cls, data: dict[str, float | bool | None]) -> Karaoke:
        return cls(
            level=data["level"],
            mono_level=data["mono_level"],
            filter_band=data["filter_band"],
            filter_width=data["filter_width"],
        )

    def __repr__(self) -> str:
        return (
            f"<Karaoke: level={self.level}, mono_level={self.mono_level}, "
            f"filter_band={self.filter_band}, filter_width={self.filter_width}>"
        )

    @property
    def level(self) -> float | None:
        return self._level

    @level.setter
    def level(self, v: float | None) -> None:
        self._level = v

    @property
    def mono_level(self) -> float | None:
        return self._mono_level

    @mono_level.setter
    def mono_level(self, v: float | None) -> None:
        self._mono_level = v

    @property
    def filter_band(self) -> float | None:
        return self._filter_band

    @filter_band.setter
    def filter_band(self, v: float | None) -> None:
        self._filter_band = v

    @property
    def filter_width(self) -> float | None:
        return self._filter_width

    @filter_width.setter
    def filter_width(self, v: float | None) -> None:
        self._filter_width = v

    @classmethod
    def default(cls) -> Karaoke:
        return cls()

    def get(self) -> dict[str, float]:
        if self.off:
            return {}
        response = {}
        if self.level is not None:
            response["level"] = self.level
        if self.mono_level is not None:
            response["monoLevel"] = self.mono_level
        if self.filter_band is not None:
            response["filterBand"] = self.filter_band
        if self.filter_width is not None:
            response["filterWidth"] = self.filter_width
        return response

    def reset(self) -> None:
        self.level = self.mono_level = self.filter_band = self.filter_width = None



from __future__ import annotations

import os

# noinspection PyProtectedMember
from pylav._internals.functions import _get_path as __get_path
from pylav.constants.node_features import SUPPORTED_SEARCHES as __SUPPORTED_SEARCHES
from pylav.logging import getLogger as __getLogger

__LOGGER = __getLogger("PyLav.Environment")

LOCAL_DEBUGGING = os.getenv("PYLAV__DEBUGGING")

POSTGRES_HOST = os.getenv("PYLAV__POSTGRES_HOST")
# noinspection SpellCheckingInspection
POSTGRES_PORT = os.getenv("PYLAV__POSTGRES_PORT")
# noinspection SpellCheckingInspection
POSTGRES_PASSWORD = os.getenv("PYLAV__POSTGRES_PASSWORD")
# noinspection SpellCheckingInspection
POSTGRES_USER = os.getenv("PYLAV__POSTGRES_USER")
# noinspection SpellCheckingInspection
POSTGRES_DATABASE = os.getenv("PYLAV__POSTGRES_DB")
POSTGRES_SOCKET = os.getenv("PYLAV__POSTGRES_SOCKET")
FALLBACK_POSTGREST_HOST = POSTGRES_HOST
if POSTGRES_SOCKET is not None:
    POSTGRES_PORT = None
    POSTGRES_HOST = POSTGRES_SOCKET
JAVA_EXECUTABLE = __get_path(envar_value) if (envar_value := os.getenv("PYLAV__JAVA_EXECUTABLE")) is not None else None
LINKED_BOT_IDS = (
    list(map(str.strip, envar_value.split("|")))
    if (envar_value := os.getenv("PYLAV__LINKED_BOT_IDS")) is not None
    else None
)
USE_BUNDLED_EXTERNAL_PYLAV_NODE = (
    bool(int(envar_value)) if (envar_value := os.getenv("PYLAV__USE_BUNDLED_EXTERNAL_PYLAV_NODE")) is not None else None
)

REDIS_FULL_ADDRESS_RESPONSE_CACHE = os.getenv("PYLAV__REDIS_FULL_ADDRESS_RESPONSE_CACHE")

EXTERNAL_UNMANAGED_HOST = os.getenv("PYLAV__EXTERNAL_UNMANAGED_HOST")
EXTERNAL_UNMANAGED_PORT = (
    int(envar_value) if (envar_value := os.getenv("PYLAV__EXTERNAL_UNMANAGED_PORT")) is not None else None
)
EXTERNAL_UNMANAGED_PASSWORD = os.getenv("PYLAV__EXTERNAL_UNMANAGED_PASSWORD")
EXTERNAL_UNMANAGED_SSL = (
    bool(int(envar_value)) if (envar_value := os.getenv("PYLAV__EXTERNAL_UNMANAGED_SSL")) is not None else None
)
EXTERNAL_UNMANAGED_NAME = os.getenv("PYLAV__EXTERNAL_UNMANAGED_NAME")

READ_CACHING_ENABLED = (
    bool(int(envar_value)) if (envar_value := os.getenv("PYLAV__READ_CACHING_ENABLED")) is not None else None
)

TASK_TIMER_UPDATE_BUNDLED_PLAYLISTS_DAYS = (
    max(int(envar_value), 1)
    if (envar_value := os.getenv("PYLAV__TASK_TIMER_UPDATE_BUNDLED_PLAYLISTS_DAYS")) is not None
    else None
)

TASK_TIMER_UPDATE_BUNDLED_EXTERNAL_PLAYLISTS_DAYS = (
    max(int(envar_value), 7)
    if (envar_value := os.getenv("PYLAV__TASK_TIMER_UPDATE_BUNDLED_EXTERNAL_PLAYLISTS_DAYS")) is not None
    else None
)


TASK_TIMER_UPDATE_EXTERNAL_PLAYLISTS_DAYS = (
    max(int(envar_value), 7)
    if (envar_value := os.getenv("PYLAV__TASK_TIMER_UPDATE_EXTERNAL_PLAYLISTS_DAYS")) is not None
    else None
)


# noinspection SpellCheckingInspection
DEFAULT_SEARCH_SOURCE = os.getenv("PYLAV__DEFAULT_SEARCH_SOURCE")
if DEFAULT_SEARCH_SOURCE is not None and DEFAULT_SEARCH_SOURCE not in __SUPPORTED_SEARCHES:
    DEFAULT_SEARCH_SOURCE = None

MANAGED_NODE_SPOTIFY_CLIENT_ID = os.getenv("PYLAV__MANAGED_NODE_SPOTIFY_CLIENT_ID")
MANAGED_NODE_SPOTIFY_CLIENT_SECRET = os.getenv("PYLAV__MANAGED_NODE_SPOTIFY_CLIENT_SECRET")
MANAGED_NODE_SPOTIFY_COUNTRY_CODE = os.getenv("PYLAV__MANAGED_NODE_SPOTIFY_COUNTRY_CODE")
MANAGED_NODE_APPLE_MUSIC_API_KEY = os.getenv("PYLAV__MANAGED_NODE_APPLE_MUSIC_API_KEY")
MANAGED_NODE_APPLE_MUSIC_COUNTRY_CODE = os.getenv("PYLAV__MANAGED_NODE_APPLE_MUSIC_COUNTRY_CODE")
MANAGED_NODE_YANDEX_MUSIC_ACCESS_TOKEN = os.getenv("PYLAV__MANAGED_NODE_YANDEX_MUSIC_ACCESS_TOKEN")
MANAGED_NODE_DEEZER_KEY = os.getenv("PYLAV__MANAGED_NODE_DEEZER_KEY")

LOCAL_TRACKS_FOLDER = os.getenv("PYLAV__LOCAL_TRACKS_FOLDER")
DATA_FOLDER = os.getenv("PYLAV__DATA_FOLDER")
ENABLE_NODE_RESUMING = (
    bool(int(envar_value)) if (envar_value := os.getenv("PYLAV__ENABLE_NODE_RESUMING")) is not None else None
)
ENABLE_NODE_RESUMING = False
# TODO:
#  - Add support for resuming nodes



from __future__ import annotations

import dataclasses

from pylav.type_hints.dict_typing import JSON_DICT_TYPE


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class State:
    time: int
    connected: bool
    ping: int
    position: int | None = 0

    def to_dict(self) -> JSON_DICT_TYPE:
        return {
            "time": self.time,
            "connected": self.connected,
            "ping": self.ping,
            "position": self.position,
        }

    def __repr__(self) -> str:
        return (
            f"<State(time={self.time} " f"position={self.position} " f"connected={self.connected} " f"ping={self.ping})"
        )
